"""Send rcon commands to server report response to raw socket

Usage:
  munerator [options] rcon

Options:
  -v --verbose         Verbose logging
  --rcon-socket url    ZMQ socket for rcon commands [default: tcp://127.0.0.1:9005]
  --oa-addr host       Host or ip where openarena runs [default: quake.brensen.com]
  --oa-port port       Port which openarena runs on [default: 1200]
  --rcon-passwd pass   Rcon password
  --raw-socket url     ZMQ socket for raw responses [default: tcp://127.0.0.1:9000]

"""
from docopt import docopt
import zmq
import logging
log = logging.getLogger(__name__)
from vendor.pyquake3 import Administrator
import time


def eventstream(zmq_socket, rcon_connection, raw_socket):
    while True:
        cmd = zmq_socket.recv_string()
        # send some commands without rcon
        if cmd in ['getstatus']:
            response_type, response = rcon_connection.command(str(cmd))
        else:
            response_type, response = rcon_connection.rcon_command(str(cmd))
        log.debug("cmd:%s, response:%s" % (cmd, response))

        now = time.time()

        # put response into translate module for parsing
        for line in response.splitlines():
            if line:
                raw_socket.send_string("%s cmd:%s response_type:%s response:%s" % (now, cmd, response_type, line))


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    zmq_socket = context.socket(zmq.PULL)
    zmq_socket.bind(args['--rcon-socket'])

    raw_socket = context.socket(zmq.PUSH)
    raw_socket.connect(args['--raw-socket'])

    rcon_connection = Administrator(args['--oa-addr'], int(args['--oa-port']), args['--rcon-passwd'])

    eventstream(zmq_socket, rcon_connection, raw_socket)



"""Change aspects about the current game

Usage:
  munerator [options] changer

Options:
  -v --verbose          Verbose logging
  --context-socket url  ZMQ socket for context events [default: tcp://quake.brensen.com:9002]
  --rcon-socket url     ZMQ socket for rcon commands [default: tcp://127.0.0.1:9005]

"""
import json
import logging
import re
import time
from functools import partial

import zmq
from docopt import docopt
from Levenshtein import ratio


log = logging.getLogger(__name__)
voting_enabled = False


class Changer(object):

    def __init__(self, in_socket, rcon_socket):
        self.in_socket = in_socket
        self.rcon_socket = rcon_socket
        self.votecache = []

    def rcon(self, cmd):
        self.rcon_socket.send_string(cmd)
        response = self.rcon_socket.recv_string()
        log.debug('cmd:%s response:%s' % (cmd, str(response)))

    def rcon_get_value(self, value):
        self.rcon_socket.send_string(value)
        try:
            response = json.loads(self.rcon_socket.recv_string())
            assert len(response) == 2
        except:
            log.error('invalid response')
            return None

        m = re.match('"%s" is:"(?P<value>[^\^]*)\^7".*' % value, response[1])
        if m:
            return m.group('value')
        else:
            return None

    def change(self):
        while True:
            kind, data = self.in_socket.recv_string().split(' ', 1)
            log.debug('got %s: %s' % (kind, data))
            data = json.loads(data)

            if kind == 'initgame':
                # reset votecache on game start
                self.votecache = []
            elif kind in ['clientbegin', 'clientdisconnect']:
                fraglimit = int(self.rcon_get_value('fraglimit'))
                num_players = data.get('game_info', {}).get('num_players')

                # increate fraglimit by 8 for every two players joining
                new_fraglimit = int(num_players) / 2 * 8
                new_fraglimit = new_fraglimit = max(10, min(new_fraglimit, 30))

                log.debug('fraglimit:%s new_fraglimit:%s' % (fraglimit, new_fraglimit))

                if new_fraglimit > fraglimit:
                    log.info('new fraglimit %s' % new_fraglimit)
                    self.rcon('set fraglimit %s' % new_fraglimit)
            elif voting_enabled and kind == 'say':
                self.voting(data)

    def voting(self, data):
        text = data.get('text')

        if ratio(text, u'instagib') > 0.75:
            change = ('gametype', 'instagib')
        elif ratio(text, u'again') > 0.75:
            change = ('map', 'restart')
        else:
            return

        client_id = data.get('client_id')
        self.record_vote(client_id, change)

        num_players = len(data.get('game_info', {}).get('clients', {}))
        votes = self.votes_for_change(change)
        log.info('votes:%s num_players:%s' % (votes, num_players))
        if votes >= num_players / 2:
            changetype, value = change
            for i in range(4):
                self.rcon('say got %s votes for %s %s, changing game!' % (votes, changetype, value))
                time.sleep(1)
            self.apply_change(change)
        else:
            votes_needed = (num_players / 2) - votes
            changetype, value = change
            self.rcon('say need %s more votes for %s %s' % (votes_needed, changetype, value))

    def apply_change(self, change):
        changetype, value = change
        if changetype == 'gametype':
            self.change_gametype(value)
        elif changetype == 'map' and value == 'restart':
            self.change_gametype('')

    def record_vote(self, client_id, change):
        vote = (client_id, change)
        self.votecache.append(vote)

    def votes_for_change(self, change):
        votes = []
        for client_id, change_vote in self.votecache:
            if change == change_vote:
                if client_id not in votes:
                    votes.append(client_id)
        return len(votes)

    def change_gametype(self, gametype):
        nextmap = self.rcon_get_value('nextmap')

        self.rcon('set g_rockets 0')
        self.rcon('set g_instantgib 0')

        if gametype == 'instagib':
            self.rcon('set g_instantgib 2')
        elif gametype == 'rockets':
            self.rcon('set g_rockets 1')

        self.rcon('map_restart')
        time.sleep(2)
        self.rcon('set nextmap %s' % nextmap)


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()

    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])

    filters = ['initgame', 'clientbegin', 'clientdisconnect', 'say']
    add_filter = partial(in_socket.setsockopt, zmq.SUBSCRIBE)
    map(add_filter, filters)

    rcon_socket = context.socket(zmq.PUSH)
    rcon_socket.connect(args['--rcon-socket'])

    c = Changer(in_socket, rcon_socket)
    c.change()



"""Store game events in database

Usage:
  munerator [options] store

Options:
  -v --verbose            Verbose logging
  --context-socket url    ZMQ socket for conext event [default: tcp://127.0.0.1:9002]
  --database ip:port      Host and port for mongo database [default: 127.0.0.1:27017]
  --rcon-socket url       ZMQ socket for rcon commands [default: tcp://127.0.0.1:9005]

"""
import json
import logging
from functools import partial

import zmq
from docopt import docopt
from munerator.common.models import Games, Players, Votes, Gamemaps
from eve import Eve
from eve_mongoengine import EveMongoengine
from mongoengine.queryset import Q

log = logging.getLogger(__name__)


def handle_events(in_socket, rcon_socket):
    """
    Loop over incoming messages and handle them individually.
    """
    # prime commands for current game info
    rcon_socket.send_string('status')
    rcon_socket.send_string('getstatus')

    log.info('listening for game events')
    while True:
        msg = in_socket.recv_string()

        log.debug('got: %s' % msg)
        data = json.loads(msg.split(' ', 1)[-1])
        kind = data.get('kind')

        try:
            handle_event(kind, data, rcon_socket)
        except:
            log.exception('error in event handling')


def handle_event(kind, data, rcon_socket):
    """
    Parse event message and update database with new information.
    """

    # get player and/or game id from data
    player_id = str(data.get('client_info', {}).get('guid', ''))
    timestamp = str(data.get('game_info', {}).get('timestamp', ''))

    player, new = Players.objects.get_or_create(guid=player_id)
    game, new = Games.objects.get_or_create(timestamp=timestamp)

    # handle player updates
    if player and kind in ['clientbegin', 'clientdisconnect', 'clientuserinfochanged', 'playerscore', 'clientstatus']:
        # on name change, store previous name
        if data['client_info'] and player.name != data['client_info']['name']:
            player.update(add_to_set__names=data['client_info']['name'])

        # update variable data
        player.update(**{'set__%s' % k: v for k, v in data['client_info'].items() if not k.endswith('id')})

        # add player to game
        if game:
            game.update(add_to_set__players=player)

        log.info('updated player')

    # handle game updates
    if game and kind in ['initgame', 'shutdowngame', 'getstatus']:
        # reset other games current status
        if kind == 'initgame':
            Games.objects(current=True).update(set__current=False)

        # update variable data
        game.update(**{'set__%s' % k: v for k, v in data['game_info'].items()if not k.endswith('id')})

        # set game map
        gamemap, new = Gamemaps.objects.get_or_create(name=data['game_info']['mapname'])
        game.update(set__gamemap=gamemap)

        if data.get('extras'):
            game.options = data.get('extras')
            game.save()

        log.info('updated game')

        # reset players online status just to be sure
        Players.objects(Q(online=True) | Q(score__ne=None) | Q(team__ne=None)).update(
            set__online=False, set__score=None, set__team=None)

    # handle votes
    if kind == 'say' and player and game:
        vote = None
        if data.get('text') in ['+1', 'gg', 'GG', 'GGG', 'like', 'fuckthismaprocks', '++', '+1337']:
            vote = Votes(player=player, game=game, vote=1)
        elif data.get('text') in ['-1', '--1', '-11', '-1000', '-2', 'dislike', 'hate', 'RAGE!!!', 'fuckthismap', '--']:
            vote = Votes(player=player, game=game, vote=-1)
        if vote:
            vote.save()
            rcon_socket.send_string('say %s^7 your vote has been counted' % player.name)
            log.info('saved vote')

        # add vote to game
        game.update(add_to_set__votes=vote)


def setup_eve_mongoengine(host, port):
    # app settings
    my_settings = {
        'MONGO_HOST': host,
        'MONGO_PORT': int(port),
        'MONGO_PASSWORD': None,
        'MONGO_USERNAME': None,
        'MONGO_DBNAME': 'munerator',
        'DOMAIN': {'eve-mongoengine': {}},
    }
    app = Eve(settings=my_settings)

    # setup eve mongodb database
    ext = EveMongoengine(app)

    # add models
    ext.add_model([Players, Games, Votes, Gamemaps])


def main(argv):
    args = docopt(__doc__, argv=argv)

    # setup zmq input socket
    context = zmq.Context()
    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])
    in_socket.setsockopt(zmq.SUBSCRIBE, '')

    # apply message filters
    filters = [
        'initgame', 'shutdowngame', 'clientdisconnect', 'say',
        'clientbegin', 'clientuserinfochanged', 'playerscore'
    ]
    add_filter = partial(in_socket.setsockopt, zmq.SUBSCRIBE)
    map(add_filter, filters)

    # setup rcon socket
    rcon_socket = context.socket(zmq.PUSH)
    rcon_socket.connect(args['--rcon-socket'])

    # setup database
    host, port = args['--database'].split(':')

    setup_eve_mongoengine(host, port)

    # start event loop
    handle_events(in_socket, rcon_socket)



"""Munerator

Usage:
  munerator [options] <command> [<args>...]

Commands:
    wrap <cmd>
    trans
    context
    ledbar
    old
    voting
    listen
    rcon
    help

Options:
  -v --verbose  Verbose logging
  -h --help     Show this screen.
  --version     Show version.

"""
from docopt import docopt

import pkg_resources
import logging
import importlib

version = pkg_resources.get_distribution("munerator").version


def main():
    args = docopt(__doc__, version=version, options_first=True)
    argv = [args['<command>']] + args['<args>']

    logging.basicConfig(level=logging.INFO)

    if '-v' in argv or '--verbose' in argv or args.get('--verbose'):
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug('debug logging enabled')

    module_name = 'munerator.' + args['<command>']
    app = importlib.import_module(module_name)
    app.main(argv)



"""Display game events on ledbar

Usage:
  munerator [options] ledbar

Options:
  -v --verbose          Verbose logging
  --context-socket url  ZMQ socket for context events [default: tcp://quake.brensen.com:9002]
  --ledbar-api url      URL to ledbar api [default: http://10.110.0.119/led/{led}/{color_code}]
  --numleds num         Number of leds in the bar to use [default: 19]

"""
import collections
import itertools
import json
import logging
import urllib2
from functools import partial

import zmq
from docopt import docopt
from webcolors import name_to_hex

log = logging.getLogger(__name__)


class Ledbar(object):

    def __init__(self, numleds, ledbar_api):
        self.prev_state = ['#000000'] * numleds
        self.ledbar_api = ledbar_api
        self.numleds = numleds

    def api_call(self, url):
        try:
            urllib2.urlopen(url, timeout=0.5)
        except:
            log.debug('urlopen failed')

    def update_leds(self, state):
        for i, color in enumerate(state):
            if not color.startswith('#'):
                color = name_to_hex(color)

            if color != self.prev_state[i]:
                led = self.numleds - (i + 1)
                color_code = color.lstrip('#')

                url = self.ledbar_api.format(led=led, color_code=color_code)
                log.debug('url: %s' % url)
                self.api_call(url)

            self.prev_state[i] = color


def update_ledbar(in_socket, numleds, ledbar):
    ids = None

    while True:
        msg = in_socket.recv_string()

        log.debug('got : %s' % msg)
        data = json.loads(msg.split(' ', 1)[-1])
        kind = data.get('kind')
        client_id = data.get('client_id', '')

        if kind == 'initgame' or ids is None:
            ids = collections.defaultdict(itertools.count().next)
            num_players = data.get('game_info', {}).get('num_players', 0)
            state = ['white'] * num_players + ['black'] * (numleds - num_players)
        elif client_id.isdigit() and int(client_id) < numleds:
            s = state[ids[int(client_id)]]
            if kind == 'clientbegin':
                team = data.get('client_info', {}).get('team')
                if team and team in ['red', 'blue']:
                    s = team
                else:
                    s = 'white'
            elif kind == 'hit':
                if s in ['cyan', 'invisible']:
                    continue
                health = int(data.get('health'))
                if health < 50:
                    s = 'orange'
                elif health < 25:
                    s = 'orangered'
            elif kind == 'quad':
                s = 'cyan'
            elif kind == 'invisible':
                s = 'black'
            elif kind == 'clientdisconnect':
                s = '#212121'
            state[ids[int(client_id)]] = s
        elif kind == 'killer':
            state[ids[int(client_id)]] = 'green'
        elif kind == 'killed':
            state[ids[int(client_id)]] = 'red'

        ledbar.update_leds(state)


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])

    filters = ['initgame', 'shutdowngame', 'clientdisconnect', 'quad',
               'clientbegin', 'clientuserinfochanged', 'hit', 'kill']
    add_filter = partial(in_socket.setsockopt, zmq.SUBSCRIBE)
    map(add_filter, filters)

    numleds = int(args['--numleds'])
    ledbar_api = args['--ledbar-api']

    ledbar = Ledbar(numleds, ledbar_api)

    update_ledbar(in_socket, numleds, ledbar)



__all__ = ['Players', 'Games', 'Votes']

from mongoengine import (BooleanField, Document, IntField, DictField,
                         ListField, ReferenceField, StringField, URLField)


class Players(Document):
    guid = StringField()
    name = StringField()
    names = ListField(StringField())
    online = BooleanField()
    score = IntField()
    team = StringField()
    team_id = IntField()
    skill = StringField()
    address = StringField()
    bot = BooleanField()


class Gamemaps(Document):
    name = StringField()
    levelshot = URLField()
    images = ListField(URLField())
    gametypes = ListField(IntField())


class Votes(Document):
    game = ReferenceField('Games')
    player = ReferenceField('Players')
    vote = IntField()


class Games(Document):
    timestamp = StringField()
    mapname = StringField()
    gamemap = ReferenceField(Gamemaps)
    gametype = IntField()
    players = ListField(ReferenceField('Players'))
    votes = ListField(ReferenceField('Votes'))
    state = StringField()
    start = StringField()
    stop = StringField()
    num_players = IntField()
    current = BooleanField()
    options = DictField()



"""Translate events

Usage:
  munerator [options] old

Options:
  -v --verbose          Verbose logging
  --context-socket url  ZMQ socket for context events [default: tcp://quake.brensen.com:9002]
  --old-api url         URL to old API [default: http://quake.ijohan.nl/events]

"""
from docopt import docopt
import zmq
import logging
log = logging.getLogger(__name__)

import urllib2
import json
import socket


def proxy_to_old_api(in_socket, base_url):
    while True:
        kind, data = in_socket.recv_string().split(' ', 1)
        log.debug('got %s: %s' % (kind, data))
        data = json.loads(data)

        if kind == 'say':
            url = '%s/say/%s/%s' % (base_url, data.get('player_name'), data.get('text'))
        elif kind == 'initgame':
            url = '%s/InitGame/%s' % (base_url, data.get('mapname'))

        log.info('calling: %s' % url)
        try:
            response = urllib2.urlopen(url, timeout=3)
            log.debug('response %s' % response)
        except socket.timeout:
            log.debug('timeout')
        except:
            log.error('failed', exc_info=True)


def main(argv):
    args = docopt(__doc__, argv=argv)

    context = zmq.Context()
    in_socket = context.socket(zmq.SUB)
    in_socket.connect(args['--context-socket'])
    in_socket.setsockopt(zmq.SUBSCRIBE, 'say')
    in_socket.setsockopt(zmq.SUBSCRIBE, 'initgame')

    proxy_to_old_api(in_socket, args['--old-api'])



from setuptools import setup, find_packages


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content

setup(
    name='munerator',
    version="0.10.6",
    description='Manager of OpenArena battles',
    long_description=read("README.rst"),
    author='Johan Bloemberg',
    author_email='mail@ijohan.nl',
    url='https://github.com/aequitas/munerator',
    install_requires=[
        'six',
        'docopt',
        'pyzmq',
        'webcolors',
        'python-Levenshtein',
        'eve==0.3',
        'flask',
        'tornado',
        'mongoengine',
        'eve-mongoengine',
        'python-dateutil'
    ],
    packages=find_packages(),
    license=read("LICENSE"),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    entry_points={
        'console_scripts': [
            "munerator = munerator:main",
        ]
    },
)
